{
{-# OPTIONS -fno-warn-incomplete-patterns -fno-warn-overlapping-patterns #-}
{-| There are three shift\/reduce conflicts:

    > \x -> x + y

    can be parsed (incorrectly) as @(\\x -> x) + y@ or
    (correctly) as @\\x -> x + y@. The parser will @shift@ and hence do the
    right thing. The same problem occurs for @`op`@ and @->@ in place of @+@.
-}
module Syntax.Parser (
    -- * Exported Functions
    parse
    , parseFile
    , parseLiterate
    , parseLiterateFile
    , programParser
    , termParser
    , tokensParser
    , interactParser
{-  REMOVE_IF_HADDOCK
    -- * Private Functions
    , mkLam
    , mkPi
    , mkInfix
    REMOVE_IF_HADDOCK -}
    ) where

import Syntax.Position
import Syntax.ParseMonad
import Syntax.Concrete
import Syntax.Lexer

import Utils.Monads

}

%name termParser Term
%name programParser Program
%name tokensParser Tokens
%name interactParser Interactive
%tokentype { Token }
%monad { Parser }
%lexer { lexer } { TkEOF }
-- %expect 3

%token
    '('	       { TkLParen $$ }
    ')'	       { TkRParen $$ }
    '\\'       { TkLambda $$ }
    '->'       { TkArrow $$ }
    '=>'       { TkDblArrow $$ }
    '_'	       { TkUnderscore $$ }
    '$'	       { TkDollar $$ }
    id	       { TkId $$ }
    int	       { TkInt $$ }
    atom       { TkAtom $$ }
    ':'	       { TkColon $$ }
    univ       { TkUniverse $$ }
    'with'     { TkWith $$ }
    'mutual'   { TkMutual $$ }
    'abstract' { TkAbstract $$ }
    'class'    { TkClass $$ }
    'instance' { TkInstance $$ }
    'let'      { TkLet $$ }
    'in'       { TkIn $$ }
    'data'     { TkData $$ }
    'rewrite'  { TkRewrite $$ }
    'extends'  { TkExtends $$ }
    'where'    { TkWhere $$ }
    Set	       { TkSet $$ }
    'ISet'     { TkISet $$ }
    'Prop'     { TkProp $$ }
    'Type'     { TkType $$ }
    'infix'    { TkInfix $$ }
    'infixl'   { TkInfixL $$ }
    'infixr'   { TkInfixR $$ }
    plugin     { TkPlugin $$ }
    'sig'      { TkSig $$ }
    'struct'   { TkStruct $$ }
    '{'	       { TkOpenBrace $$ }
    '}'	       { TkCloseBrace $$ }
    '['	       { TkOpenSquare $$ }
    ']'	       { TkCloseSquare $$ }
    'do'       { TkDo $$ }
    '<-'       { TkFrom $$ }
    vopen      { TkVOpenBrace $$ }
    vclose     { TkVCloseBrace $$ }
    '='	       { TkEqual $$ }
    ';'	       { TkSemi $$ }
    vsemi      { TkVSemi $$ }
    ','	       { TkComma $$ }
    '`'	       { TkBackQuote $$ }
    '.'	       { TkDot $$ }
    '?'	       { TkOp ($$,"?") }
    op	       { TkOp $$ }
    TeX	       { TkTeX $$ }

%%

-- Tokens

Token
    : '('	 { TkLParen $1 }
    | ')'	 { TkRParen $1 }
    | '\\'	 { TkLambda $1 }
    | '->'	 { TkArrow $1 }
    | '=>'	 { TkDblArrow $1 }
    | '_'	 { TkUnderscore $1 }
    | '$'	 { TkDollar $1 }
    | id	 { TkId $1 }
    | atom	 { TkAtom $1 }
    | int	 { TkInt $1 }
    | ':'	 { TkColon $1 }
    | univ	 { TkUniverse $1 }
    | 'with'	 { TkWith $1 }
    | 'mutual'	 { TkMutual $1 }
    | 'abstract' { TkAbstract $1 }
    | 'class'	 { TkClass $1 }
    | 'instance' { TkInstance $1 }
    | 'let'	 { TkLet $1 }
    | 'in'	 { TkIn $1 }
    | 'data'	 { TkData $1 }
    | 'rewrite'  { TkRewrite $1 }
    | 'extends'	 { TkExtends $1 }
    | 'where'	 { TkWhere $1 }
    | Set	 { TkSet $1 }
    | 'ISet'	 { TkISet $1 }
    | 'Prop'	 { TkProp $1 }
    | 'Type'	 { TkType $1 }
    | 'infix'	 { TkInfix $1 }
    | 'infixl'	 { TkInfixL $1 }
    | 'infixr'	 { TkInfixR $1 }
    | plugin	 { TkPlugin $1 }
    | 'sig'	 { TkSig $1 }
    | 'struct'	 { TkStruct $1 }
    | 'do'	 { TkDo $1 }
    | '<-'	 { TkFrom $1 }
    | '{'	 { TkOpenBrace $1 }
    | '}'	 { TkCloseBrace $1 }
    | '['	 { TkOpenSquare $1 }
    | ']'	 { TkCloseSquare $1 }
    | vopen	 { TkVOpenBrace $1 }
    | vclose	 { TkVCloseBrace $1 }
    | '='	 { TkEqual $1 }
    | ';'	 { TkSemi $1 }
    | vsemi	 { TkVSemi $1 }
    | ','	 { TkComma $1 }
    | '`'	 { TkBackQuote $1 }
    | '.'	 { TkDot $1 }
    | '?'	 { TkOp ($1,"?") }
    | op	 { TkOp $1 }
    | TeX	 { TkTeX $1 }

Tokens	: Token Tokens		{ $1 : $2 }
	|			{ [] }

-- Interactive
Interactive :: { Either Term Typing }
Interactive : Term	    { Left $1 }
	    | Term ':' Term { Right (Typing $1 $3) }

-- Programs

Program : Init Tops		{ $2 }

-- Definitions

Top :	topen MaybeDefs close	{ $2 }

Tops
    : Top		{ $1 }
    | Top TeX Tops	{ $1 ++ $3 }
    | TeX Tops		{ $2 }

topen :				{% pushCurrentContext }

Definitions
    : vopen MaybeDefs close	{ $2 }

close	: vclose	{ () }
	| error		{% popContext }

semi	: ';'		{ $1 }
	| vsemi		{ $1 }

-- Definitions must be separated by semicolons; however, we are quite liberal.
-- For instance, they can be separated by multiple semicolons, and trailing
-- semicolons after the last definition are also allowed.

-- Not anymore, we aren't. (Not necessary with layout?)

MaybeDefs
    :			    { [] }
    | Def		    { [$1] }
    | Def semi MaybeDefs    { $1:$3 }

With : '=' Term				{ NoWith $2 }
     | 'with' CommaTerms '=>' Clauses   { With $2 $4 }

Clauses
    : vopen Clauses0 close  { $2 }

Clauses0
    : {- empty -}   { [] }
    | Clauses1	    { $1 }

Clauses1
    : Clause		    { [$1] }
    | Clause semi Clauses1  { $1 : $3 }

Clause
    : HPatterns Patterns With	{ Clause (fuseRange $1 ($1 ++ $2)) $1 $2 $3 }

Def : Var HPatterns1 Patterns With
			{ mkFunClause	(fuseRange $1 $4) $1 $2 $3 $4 }
    | Var Patterns With
			{ mkFunClause	(fuseRange $1 $3) $1 [] $2 $3 }
    | Pattern1 InfixOp Pattern1 With
			{ mkFunClause	(getRange $2)    -- the range says where the defined name is
					$2 [] [$1,$3] $4
			}
    | Axiom				    { $1 }
    | 'data' Var HTelescope Telescope ':' Term0 'where'
	Constructors			    { mkData False $1 $2 $3 $4 $6 $8 }
    | 'sig' Var HTelescope Telescope ':' Term0 Extends 'where'
	Constructors			    { Signature (fuseRange $1 $8)
						$2 $3 $4 $6 $7 $9
					    }
    | 'infix' int CommaOps		    { mkInfix NonAssoc $1 $2 $3 }
    | 'infixl' int CommaOps		    { mkInfix LeftAssoc $1 $2 $3 }
    | 'infixr' int CommaOps		    { mkInfix RightAssoc $1 $2 $3 }
    | 'rewrite' RewriteTerms		    { Rewrite (fuseRange $1 $2) $2 }
    | 'mutual' Definitions		    { Mutual (fuseRange $1 $2) $2 }
    | 'abstract' Definitions		    { Abstract (fuseRange $1 $2) $2 }

Extends
    : 'extends' CommaNames  { $2 }
    | {- empty -}	    { [] }

Axiom
    : Var Telescope ':' Term	{ TypeSig (fuseRange $1 $4) $1 $2 $4 }

Where
    :					    { [] }
    | 'where' LocalDefs			    { $2 }

Constructors
    : vopen Constrs close		    { $2 }
-- As definitions a ^^^^^ is required instead of vclose

Constrs1
    : Constr semi Constrs1  { $1 ++ $3 }
    | Constr		    { $1 }

Constrs
    : Constrs1		    { $1 }
    | {- empty -}	    { [] }

Constr
    : CommaNames Telescope ':' Term { mkConstrs $1 $2 $4 }

Patterns
    : Patterns1	    { $1 }
    | {- empty -}   { [] }

Patterns1
    : Pattern Patterns1	    { $1 : $2 }
    | Pattern		    { [$1] }

HPatterns
    : {- empty -}		{ [] }
    | HPatterns1		{ $1 }

HPatterns1
    : '{' Pattern0 '}'		 { [$2] }
    | '{' Pattern0 '}' HPatterns1 { $2 : $4 }

Pattern0
    : Pattern0 InfixOp Pattern1	{ InfixConP $1 $2 $3 }
    | Pattern1			{ $1 }

Pattern1
    : Var HPatterns1 Patterns1	{ mkConP (getRange ($1,$2 ++ $3))
					 $1 $2 $3
				}
    | Var Patterns1		{ mkConP (getRange ($1,$2))
					 $1 [] $2
				}
    | Var HPatterns1		{ mkConP (getRange ($1,$2))
					 $1 $2 []
				}
    | Pattern			{ $1 }

Pattern
    : int			    { uncurry mkNatP $1 }
    | Var			    { VarP $1 }
    | '_'			    { WildP $1 }
    | '[' MaybePatList ']'	    { mkListP (fuseRange $1 $3) $2 }
    | '(' Pattern0 ')'		    { $2 }
    | '(' Pattern0 ',' PatList ')'  { mkPairP (fuseRange $1 $5) ($2 : $4) }

TermLamPattern : Term5	{% mkLamPat $1 }

-- We allow a restricted form of patterns [ x | (p,p) ] in Pis and Lambdas.
LamPattern
    : Var		    { VarP $1 }
    | '_'		    { WildP $1 }
    | '(' CommaLamPats ')'  { mkPairP (fuseRange $1 $3) $2 }

CommaLamPats
    : LamPattern		    { [$1] }
    | LamPattern ',' CommaLamPats   { $1 : $3 }

SpaceLamPats0
    : {- empty -}	{ [] }
    | SpaceLamPats	{ $1 }

HiddenLamPats
    : {- empty -}			{ [] }
    | '{' LamPattern '}' HiddenLamPats	{ $2 : $4 }

SpaceLamPats
    : LamPattern		{ [$1] }
    | LamPattern SpaceLamPats	{ $1 : $2 }

LocalDefs : Definitions	{ $1 }

RewriteTerms
    : vopen SemiTerms close { $2 }

-- Telescope
Telescope
    : Telescopes    { $1 }
    | {- empty -}   { [] }

HTelescope
    : HTelescopes   { $1 }
    | {- empty -}   { [] }

Telescopes
    : '(' PatList ':' Term ')' Telescopes {% mkTelescope $2 $4 $6 }
    | '(' PatList ':' Term ')'            {% mkTelescope $2 $4 [] }

HTelescopes
    : '{' PatList ':' Term '}' HTelescopes  {% mkTelescope $2 $4 $6 }
    | '{' PatList ':' Term '}'		    {% mkTelescope $2 $4 [] }

SigmaEntry :: { ([Name], Term) }
SigmaEntry
    : CommaTermNames ':' Term	    { ($1,$3) }
    | Term			    { ([dontCare],$1) }

SigmaTel0
    : SigmaEntry ';' SigmaTel0	{ $1 : $3 }
    | Term			{ [([dontCare],$1)] }

SigmaTel
    : SigmaEntry ';' SigmaTel0	{ $1 : $3 }

StructBody
    : StructEntry semi StructBody   { $1 : $3 }
    | StructEntry		    { [$1] }

StructEntry :
    Var HiddenLamPats SpaceLamPats0 '=' Term	{ ($1, $2, $3, $5) }

Statements
    : vopen Statements1 close	{ $2 }

Statements1
    : Term			 { [SNoBind $1] }
    | Statement semi Statements1 { $1 : $3 }

Statement
    : Term			{ SNoBind $1 }
    | TermLamPattern '<-' Term  { SBind $2 $1 $3 }
    | 'let' LocalDefs		{ SLet (fuseRange $1 $2) $2 }

Projection
    : int   {% mkProj $1 }
    | Var   { Field $1 }

-- Terms

Term : Term0 Where  { mkWhere $1 $2 }

Term0
	-- should really be a list of vars, but we have to
	-- parse it as a TermList to avoid conflicts with
	-- the pair syntax
    : '(' TermList ':' Term ')' '->' Term0
				    {% mkPi (fuseRange $1 $7)
					    $2 $4 $7
				    }
    | Term1 '->' Term0		    { Pi (fuseRange $1 $3)
					 (WildP (getRange $1)) $1 $3 }
    | Term1			    { $1 }

Term1
    : Term1b '$' Term1 { App $1 $3 }
    | Term1b	       { $1 }

Term1b
    : Term1b InfixOp Term2		{ InfixApp $1 $2 $3 }
    | Term2				{ $1 }

Term2
    : '\\' SpaceLamPats '->' Term0  { mkLam (fuseRange $1 $4) $2 $4 }
    | 'let' LocalDefs 'in' Term0	    { Let (fuseRange $1 $4) $2 $4 }
    | Term4 '.' 'do' Statements	    { mkDo $1 $4 }
    | 'rewrite' RewriteTerms 'in' Term0 { RewriteIn (fuseRange $1 $4) $2 $4 }
    | Term3			    { $1 }

Term3
    : Term3 Term4		     { App $1 $2 }
    | Term3 '{' Term '}'	     { HApp $1 $3 }
    | Term4			     { $1 }

Term4
    : Term4 '.' Projection  { Project (fuseRange $1 $3) $1 $3 }
    | Term5		    { $1 }

-- Atomic Terms

Term5
    : Var				{ Id $1 }
    | int				{ uncurry mkNat $1 }
    | Universe				{ $1 }
    | plugin '(' MaybeTermList ')'	{ CallPlugin (fuseRange (fst $1) $4) (snd $1) $3 }
    | '?'				{ CallPlugin $1 "interact" [] }
    | '_'				{ Meta $1 }
    | '(' Term ')'			{ Paren (fuseRange $1 $3) $2 }
    | '[' MaybeTermList ']'		{ mkList (fuseRange $1 $3) $2 }
    | '{' SigmaTel '}'			{% mkSigma (fuseRange $1 $3) $2 }
    | '(' Term ',' TermList ')'		{ mkPair (fuseRange $1 $5) ($2 : $4) }
    | 'struct' vopen StructBody close	{ Struct (fuseRange $1 $3) $3 }

CommaTerms
    : Term0		    { [$1] }
    | Term0 ',' CommaTerms  { $1 : $3 }

SemiTerms
    : Term0		    { [$1] }
    | Term0 semi SemiTerms  { $1 : $3 }

Universe
    : univ				    {% mkUniv $1 }
    | 'ISet'				    { ISet $1 }
    | Set				    {% mkSet $1 }
    | 'Prop'				    {% mkProp $1 }
    | 'Type'				    {% mkType $1 }

InfixOp
    : '`' id '`'			    { mkName $2 }
    | op				    { mkName $1 }

MaybeTermList
    :                                       { [] }
    | TermList                              { $1 }

TermList
    : Term                 { [$1] }
    | Term ',' TermList    { $1 : $3 }

MaybePatList
    :                                   { [] }
    | PatList                           { $1 }

PatList
    : Pattern0				{ [$1] }
    | Pattern0 ',' PatList		{ $1 : $3 }

-- Names

CommaNames
    : Var		    { [$1] }
    | Var ',' CommaNames    { $1 : $3 }

CommaTermNames
    : Term5			{% do x <- isName $1; return [x] }
    | Term5 ',' CommaTermNames  {% do x <- isName $1; return (x : $3) }

CommaOps
    : InfixOp		    { [$1] }
    | InfixOp ',' CommaOps  { $1 : $3 }

Var : id				    { mkName $1 }
    | atom				    { let (r,x) = $1 in mkName (r,"\"" ++ x ++ "\"") }
    | '(' op ')'			    { mkName $2 }

-- Name : id   { mkName $1 }
--      | op   { mkName $1 }
--      | atom { let (r,x) = $1 in mkName (r,"\"" ++ x ++ "\"") }

Init : { }

{

-- Parsing

parse :: Parser a -> String -> ParseResult a
parse p = parse' defaultParseFlags (startPos "") 0 p

parseFile :: Parser a -> FilePath -> IO (ParseResult a)
parseFile p = parseFile' defaultParseFlags 0 p

parseLiterate :: Parser a -> String -> ParseResult a
parseLiterate p = parse' defaultParseFlags (startPos "") tex p

parseLiterateFile :: Parser a -> FilePath -> IO (ParseResult a)
parseLiterateFile p = parseFile' defaultParseFlags tex p

programParser	:: Parser Program
termParser	:: Parser Term
tokensParser	:: Parser [Token]

happyError = fail "Parse error"

isName :: Term -> Parser Name
isName (Id x) = return x
isName _      = fail "Expected a name, found a term"

mkProj (r,n)
    | n >= 1	= return $ Proj r n
    | otherwise	= fail "Non-positive projection"

-- Do syntax
data Statement = SBind Range Pattern Term
	       | SLet Range [Decl]
	       | SNoBind Term

bind :: Range -> Term -> Term
bind r e = Project r e $ Field $ mkName (r,"bind")

mkDo :: Term -> [Statement] -> Term
mkDo mon [SNoBind e] = e
mkDo mon (SNoBind e : ss) = App (bind r mon) e `App` Lam r (WildP r) (mkDo mon ss)
    where
	r = getRange e
mkDo mon (SBind r p e : ss) = App (bind r mon) e `App` Lam r' p (mkDo mon ss)
    where
	r' = fuseRange r p
mkDo mon (SLet r ds : ss) = Let r ds $ mkDo mon ss

-- Sorts
mkUniv (r,n) = return $ Hash r $ Univ n
mkSet r	     = return $ Hash r $ Univ 0

mkProp r = return $ Prop r

mkType r = return $ Hash r $ Univ 1

-- | Even though this function is not exported, I want it to show up
--   in the documentation.
mkLam :: Range -> [Pattern] -> Term -> Term
mkLam r vs t = foldr (Lam r) t vs

mkPi :: Range -> [Term] -> Term -> Term -> Parser Term
mkPi r vs t u
    | all isLamPat vs	= return $ foldr (\v -> Pi r v t) u $ map getLamPat vs
    | otherwise		= happyError
    where
	isLamPat (Id _)		= True
	isLamPat (Pair _ p q)	= isLamPat p && isLamPat q
	isLamPat _		= False

	getLamPat (Id DontCare) = WildP r
	getLamPat (Id x)	= VarP x
	getLamPat (Pair r p q)	= PairP r (getLamPat p) (getLamPat q)

mkSig :: Range -> [([Name], Term)] -> Parser Term
mkSig _ _ = fail $ "Signature support is currently closed for maintenance"

mkSigma :: Range -> [([Name], Term)] -> Parser Term
mkSigma r tel = return $ sigma flat
    where
	flat = concatMap (\ (xs,t) -> [ (x,t) | x <- xs ] ) tel

	sigma [(x,t)]	  = t
	sigma ((x,t):tel) = Sigma r x t $ sigma tel

mkPair :: Range -> [Term] -> Term
mkPair r ts = foldr1 (Pair r) ts

mkPairP :: Range -> [Pattern] -> Pattern
mkPairP r ps = foldr1 (PairP r) ps

mkInfix :: Assoc -> Range -> (Range, Precedence) -> [Name] -> Decl
mkInfix ass r n vs =
    Infix   (fuseRange r (last vs))
	    ass (snd n) vs

mkTelescope :: [Pattern] -> Term -> Telescope -> Parser Telescope
mkTelescope ps t tel =
    do	xs <- mapM isVar ps
	return $ map (\x -> (x,t)) xs ++ tel
    where
	isVar (VarP x)	= return x
	isVar p		= fail "Patterns are not allowed in telescopes"

mkFunClause :: Range -> Name -> [Pattern] -> [Pattern] -> WithClause -> Decl
mkFunClause r x hs ps w = FunClause r x $ Clause r hs ps w

mkWhere :: Term -> [Decl] -> Term
mkWhere e [] = e
mkWhere e ds = Let r ds e
    where
	r = fuseRange (head ds) (last ds)

-- mkRHS :: Term -> [Decl] -> WithClause
-- mkRHS e []  = NoWith e
-- mkRHS e ds  = NoWith (Let r' ds e)

mkData sized r x htel tel t [] = Data sized (fuseRange r t) x htel tel t []
mkData sized r x htel tel t cs = Data sized (fuseRange r (last cs)) x htel tel t cs

-- | Hard coded names for zero and suc!
mkNat :: Range -> Int -> Term
mkNat r 0 = Id $ mkName (r,"zero")
mkNat r n = App (Id $ mkName (r,"suc")) (mkNat r (n - 1))

mkNatP :: Range -> Int -> Pattern
mkNatP r 0 = VarP $ mkName (r,"zero")
mkNatP r n = ConP r (mkName (r,"suc")) [] [mkNatP r (n - 1)]

mkConP :: Range -> Name -> [Pattern] -> [Pattern] -> Pattern
mkConP _ x [] [] = VarP x
mkConP r x hs ps = ConP r x hs ps

mkList :: Range -> [Term] -> Term
mkList r []	= Id $ mkName (r,"nil")
mkList r (x:xs) = Id (mkName (r,"::")) `App` x `App` mkList r xs

mkListP :: Range -> [Pattern] -> Pattern
mkListP r []	    = VarP $ mkName (r,"nil")
mkListP r (x:xs)    = ConP r (mkName (r,"::")) [] [x, mkListP r xs]

mkName :: (Range, String) -> Name
mkName (r,x) = Name r x

mkConstrs :: [Name] -> Telescope -> Term -> [Constructor]
mkConstrs xs tel t = map (\x -> Constructor (fuseRange x t) x tel t) xs

mkLamPat :: Term -> Parser Pattern
mkLamPat (Pair r p q)	= PairP r <$> mkLamPat p <*> mkLamPat q
mkLamPat (Meta r)	= return $ WildP r
mkLamPat (Id x)		= return $ VarP x
mkLamPat e		= fail $ "Invalid irrefutable pattern."

}
