{-# OPTIONS -fglasgow-exts -fallow-overlapping-instances #-}
module TypeChecker.Abstractions where

import Control.Monad.Trans
import Data.Maybe
import System.Time

import Syntax.Internal
import Syntax.Values
import Utils.Monads

import Utils.Lists

-- | Each element in the list has one more free variable than its predecessor.
newtype TeleList a = TeleList [a]

newtype DeBruijnIndex = DeBr { unDeBr :: Int }

{------------------------------------------------------------------------
    Raising a value
 ------------------------------------------------------------------------}

class Raiseable a where
    raiseByFrom	:: Int -> Int -> a -> a

raise :: Raiseable a => a -> a
raise = raiseBy 1

raiseBy	:: Raiseable a => Int -> a -> a
raiseBy n = raiseByFrom n 0

instance Raiseable a => Raiseable [a] where
    raiseByFrom k n = map (raiseByFrom k n)

instance Raiseable Value where
    raiseByFrom k n v =
	case vView v of
	    VBId m	-> vBId (unDeBr $ rbf n $ DeBr m)
	    VApp v w	-> vApp (rbf n v) (rbf n w)
	    VPi v f	-> vPi (rbf n v) (rbf n f)
	    VLam f	-> vLam (rbf n f)
	    VPair v w	-> vPair (rbf n v) (rbf n w)
	    VProj v p	-> vProj (rbf n v) p
	    VSigma v f	-> vSigma (rbf n v) (rbf n f)
	    VId _	-> v
	    VTop	-> v
	    VDef _ _	-> v
	    VCon _ _ _	-> v
	    VSort _	-> v
	    VMeta _	-> v
	    VStruct xs	-> vStruct (rbf n xs)
	where
	    rbf n x = raiseByFrom k n x

instance Raiseable ProjectionValue where
    raiseByFrom _ _ p = p

instance Raiseable DefValue where
    raiseByFrom k n (DefVal h ar fv cs t) =
	DefVal h ar fv (raiseByFrom k n cs) (raiseByFrom k n t)

instance Raiseable ActualDefinition where
    raiseByFrom k n def = mapClauses (raiseByFrom k n) def

instance Raiseable ClauseValue where
    raiseByFrom k n (ClauseVal ps v) = ClauseVal ps $ raiseByFrom k n v

instance Raiseable Type where
    raiseByFrom k n t = typeMap (raiseByFrom k n) t

instance Raiseable a => Raiseable (Abs a) where
    raiseByFrom k n = absMap (raiseByFrom k (n + 1))

instance Raiseable a => Raiseable (TeleList a) where
    raiseByFrom k n (TeleList []) = TeleList []
    raiseByFrom k n (TeleList (x:xs)) =
	    TeleList (raiseByFrom k n x : xs')
	where
	    TeleList xs' = raiseByFrom k (n + 1) (TeleList xs)

instance Raiseable String where
    raiseByFrom _ _ x = x

instance Raiseable Name where
    raiseByFrom _ _ x = x

instance Raiseable Int where
    raiseByFrom _ _ n = n

instance (Raiseable a, Raiseable b) => Raiseable (a,b) where
    raiseByFrom k n (x,y) = (raiseByFrom k n x, raiseByFrom k n y)

instance Raiseable DeBruijnIndex where
    raiseByFrom k n (DeBr m)
	| m >= n    = DeBr (m + k)
	| otherwise = DeBr m

instance (Raiseable a, Raiseable b) => Raiseable (Either a b) where
    raiseByFrom k n (Left x)  = Left $ raiseByFrom k n x
    raiseByFrom k n (Right y) = Right $ raiseByFrom k n y

{------------------------------------------------------------------------
    Instantiation
 ------------------------------------------------------------------------}

class Instantiate a where
    instantiateUnder :: Int -> a -> [Value] -> a

instantiate :: Instantiate a => a -> [Value] -> a
instantiate = instantiateUnder 0

-- Invariant: @vs@ is always valid, @n@ is the number of closures we've gone
--		under.
-- Metas can't be instantiated to variables so we don't have to look inside them.
instance Instantiate Value where
    instantiateUnder n v vs =
	case vView v of
	    VBId m
		| m < n	    -> vBId m
		| otherwise -> maybe (vBId (m - length vs)) id (vs !? m - n)
	    VApp v w	    -> vApply (inst v) (inst w)
	    VPi v f	    -> vPi (inst v) (inst f)
	    VLam f	    -> vLam (inst f)
	    VPair v w	    -> vPair (inst v) (inst w)
	    VProj v p	    -> vProject (inst v) p
	    VSigma v f	    -> vSigma (inst v) (inst f)
	    VStruct xs	    -> vStruct (inst xs)
	    VTop	    -> v
	    VId _	    -> v
	    VDef _ _	    -> v
	    VCon _ _ _	    -> v
	    VSort _	    -> v
	    VMeta _	    -> v
	where
	    inst x = instantiateUnder n x vs

instance Instantiate Type where
    instantiateUnder n t vs = typeMap (\v -> instantiateUnder n v vs) t

instance Instantiate a => Instantiate (Abs a) where
    instantiateUnder n f vs =
	absMap (\v -> instantiateUnder (n + 1) v $ raise vs) f

instance Instantiate a => Instantiate (TeleList a) where
    instantiateUnder n (TeleList xs) vs =
	case xs of
	    []	    -> TeleList []
	    x:xs    -> TeleList $ instantiateUnder n x vs : xs'
		where TeleList xs' = instantiateUnder (n + 1) (TeleList xs) (raise vs)

instance Instantiate DefValue where
    instantiateUnder n (DefVal h ar fv cs t) vs =
	uncurry (DefVal h ar fv) $ instantiateUnder n (cs,t) vs

instance Instantiate ActualDefinition where
    instantiateUnder n def vs = mapClauses (\cl -> instantiateUnder n cl vs) def

instance Instantiate ClauseValue where
    instantiateUnder n (ClauseVal ps v) vs = ClauseVal ps $ instantiateUnder n v vs

instance Instantiate Int where
    instantiateUnder _ n _ = n

instance Instantiate Name where
    instantiateUnder _ x _ = x

instance Instantiate String where
    instantiateUnder _ s _ = s

instance (Instantiate a, Instantiate b) => Instantiate (a,b) where
    instantiateUnder n (x,y) vs =
	(instantiateUnder n x vs, instantiateUnder n y vs)

instance Instantiate a => Instantiate [a] where
    instantiateUnder n xs vs = map (\x -> instantiateUnder n x vs) xs

{------------------------------------------------------------------------
    Abstraction application
 ------------------------------------------------------------------------}

absApp :: Instantiate a => Abs a -> Value -> a
absApp a w = instantiate (absBody a) [w]

{--------------------------------------------------------------------------
    Beta rules
 --------------------------------------------------------------------------}

vApps :: Value -> [Value] -> Value
vApps = foldl vApply

vApply :: Value -> Value -> Value
vApply v w =
    case vView v of
	VLam f  -> absApp f w
	VTop    -> vTop
	_	-> vApp v w

vProject :: Value -> ProjectionValue -> Value
vProject v p = vp (vView v) p
    where
	vp VTop	 p		     = vTop
	vp (VPair u v) VFst	     = u
	vp (VPair u v) VSnd	     = v
	vp (VStruct xs) (VField _ x) =
	    case Prelude.lookup x xs of
		Nothing	    -> error $ "vProject (" ++ showVal v ++ ") " ++ x
		Just (_,v)  -> v
	vp _ p			    = v `vProj` p

vProjs :: Value -> [ProjectionValue] -> Value
vProjs = foldl vProject

vAppOrProjs :: Value -> [AppOrProj] -> Value
vAppOrProjs = foldl vAppOrProj

vAppOrProj :: Value -> AppOrProj -> Value
vAppOrProj v (IsApp w)	= vApply v w
vAppOrProj v (IsProj p)	= vProject v p

