{-# OPTIONS -fglasgow-exts -fallow-overlapping-instances #-}

module TypeChecker.MetaVars where

import Control.Monad
import Control.Monad.Trans
import Data.IORef
import Data.List

import Syntax.Internal
import Syntax.Values
import Syntax.Pretty

import TypeChecker.Monad
import TypeChecker.Monad.Context
import TypeChecker.Abstractions
import TypeChecker.Values
import TypeChecker.Values.Refresh
import TypeChecker.Context
import TypeChecker.UnfoldConstants
import TypeChecker.Universes
import TypeChecker.Monad.Errors
import TypeChecker.Monad.MetaVars
import TypeChecker.Errors

import Utils.Names
import Utils.Monads

createClosedMeta :: TC Value
createClosedMeta = liftIO $ vMeta <$> newIORef Open

createMeta :: TC Value
createMeta =
    do	n <- nofBoundVars <$> getContext
	let ps = reverse $ map vBId [0 .. n - 1]
	v <- createClosedMeta
	ifM isFirstOrderUnification
	    (return v)
	    (return $ foldl vApp v ps)

validParameters :: [Value] -> Bool
validParameters = all (isBId . vView)
    where
	isBId (VBId _)	= True
	isBId _		= False

instantiateToPi :: MetaRef -> SortValue -> [Value] -> TC ()
instantiateToPi m s vs =
    do	let n	= length vs
	    ps	= map vBId [n - 1, n - 2 .. 0]
	    ps'	= vBId n : ps
	m1 <- el s . flip (foldl vApp) ps  <$> createClosedMeta
	m2 <- el s . flip (foldl vApp) ps' <$> createClosedMeta
	let body = vPi m1 (mkAbs "x" m2)
	m =: lamAbstraction n body

instantiateToSigma :: MetaRef -> SortValue -> [Value] -> TC ()
instantiateToSigma m s vs =
    do	let n	= length vs
	    ps	= map vBId [n - 1, n - 2 .. 0]
	    ps'	= vBId n : ps
	m1 <- el s . flip (foldl vApp) ps  <$> createClosedMeta
	m2 <- el s . flip (foldl vApp) ps' <$> createClosedMeta
	let body = vSigma m1 (mkAbs "x" m2)
	m =: lamAbstraction n body

-- | Restricts a set of parameters.
--   Given @? qs@, @qs' = refineParameters ps qs@ is the sequence of parameters
--   making @qs'@ refer to the parameters in @qs@ occurring in @ps@ when saying
--   @? := \xs -> ..qs'..@.
refineParameters :: [Value] -> [Value] -> Maybe [Value]
refineParameters ps qs
    | validParameters qs =
	Just $
	    do	(n, q) <- reverse $ zip [0..] $ reverse qs
		unless (any (sameBId q) ps) []
		return $ vBId n
    | otherwise		 = Nothing

sameBId v w = same (vView v) (vView w)
    where
	same (VBId n) (VBId m)	= n == m
	same _ _		= False

-- | Precondition: the meta is uninstantiated
(=:) :: MetaRef -> Value -> TC ()
m =: v =
    do	Open <- liftIO $ readIORef m
	liftIO $ writeIORef m (Instantiated v)
	wake m

-- | Check that all rigid variables are in a given set and remove any
--   flexible variables that aren't. Also check that there are no occurrences
--   of the given meta variable.
occursCheck :: MetaRef -> [Value] -> Value -> TC ()
occursCheck m xs v' =
    refreshTop v' >>= \v ->
    case vAppView v of
	VAVId x _	-> failure $ customError $ "occursCheck: unexpected named variable: " ++ show x

	VAVBId n vs	->
	    do	unless (any (sameBId $ vBId n) xs)
		    $ failure $ customError $ "rigid variable error: " ++ show xs ++ " " ++ show v
		mapM_ check vs

	VAVMeta m' vs	->
	    do	when (m == m') $ failure $ customError "occurs check"
		vs' <- case refineParameters xs vs of
			Nothing	-> failure $ customError "invalid parameters" -- TODO: postpone
			Just zs	-> return zs
		-- metas are top level so we can ignore the type
		w <- createClosedMeta
		m' =: lamAbstraction (length vs) (vApps w vs')

	VAVDef _ _ vs	    -> mapM_ check vs
	VAVCon _ _ _ vs	    -> mapM_ check vs
	VAVPi v f	    ->
	    do	check $ unEl v
		checkAbsT f
	VAVLam f	    -> checkAbsV f
	VAVSort s	    -> return ()
	VAVTop		    -> return ()
	VAVPair v w	    -> mapM_ check [v,w]
	VAVSigma w f   ->
	    do	check $ unEl w
		checkAbsT f
	VAVProj v _ vs	    ->
	    mapM_ check (v:vs)
	VAVStruct bs	    -> mapM_ check $ map (snd . snd) bs
    where
	check = occursCheck m xs
	checkAbs h f = occursCheck m (vBId 0 : raise xs) (h $ absBody f)
	checkAbsT f = checkAbs unEl f
	checkAbsV f = checkAbs id f

